from __future__ import annotations

import typing

from tugboat.constraints import require_all, require_exactly_one, require_non_empty
from tugboat.core import hookimpl
from tugboat.references import get_template_context
from tugboat.utils import check_model_fields_references, prepend_loc

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    from tugboat.references import Context
    from tugboat.schemas import (
        ContainerNode,
        ContainerTemplate,
        ScriptTemplate,
        Template,
        Workflow,
        WorkflowTemplate,
    )
    from tugboat.types import Diagnosis


@hookimpl
def analyze_template(
    template: Template, workflow: Workflow | WorkflowTemplate
) -> Iterable[Diagnosis]:
    """
    Perform the analysis of the container-based template:

    - Container template
    - Container set template
    - Script template

    This implementation only checks the specialized constraints for the
    container-based templates. The general constraints are checked by the
    generic template analyzer.
    """
    if not template.container and not template.script and not template.containerSet:
        return

    yield from check_input_artifacts(template)
    yield from check_output_parameters(template)
    yield from check_output_artifacts(template)

    ctx = get_template_context(workflow, template)
    if template.container:
        yield from prepend_loc(
            ("container",), check_shared_fields(template, ctx, template.container)
        )
    if template.script:
        yield from prepend_loc(
            ("script",), check_shared_fields(template, ctx, template.script)
        )
    if template.containerSet:
        for i, container in enumerate(template.containerSet.containers or ()):
            yield from prepend_loc(
                ("containerSet", "containers", i),
                check_shared_fields(template, ctx, container),
            )


def check_input_artifacts(template: Template) -> Iterable[Diagnosis]:
    if not template.inputs:
        return

    for idx, artifact in enumerate(template.inputs.artifacts or ()):
        yield from require_non_empty(
            model=artifact,
            loc=("inputs", "artifacts", idx),
            fields=["path"],
        )


def check_output_parameters(template: Template) -> Iterable[Diagnosis]:
    if not template.outputs:
        return

    for idx, parameter in enumerate(template.outputs.parameters or ()):
        loc = ("outputs", "parameters", idx)
        yield from require_all(
            model=parameter,
            loc=loc,
            fields=["valueFrom"],
        )

        if parameter.valueFrom:
            yield from require_non_empty(
                model=parameter.valueFrom,
                loc=(*loc, "valueFrom"),
                fields=["path"],
            )


def check_output_artifacts(template: Template) -> Iterable[Diagnosis]:
    if not template.outputs:
        return

    for idx, artifact in enumerate(template.outputs.artifacts or ()):
        yield from require_non_empty(
            model=artifact,
            loc=("outputs", "artifacts", idx),
            fields=["path"],
        )


def check_shared_fields(
    template: Template,
    context: Context,
    node: ContainerNode | ContainerTemplate | ScriptTemplate,
) -> Iterable[Diagnosis]:
    # field `env`
    for i, envvar in enumerate(node.env or ()):
        yield from require_non_empty(
            model=envvar,
            loc=("env", i),
            fields=["name"],
        )
        yield from require_exactly_one(
            model=envvar,
            loc=("env", i),
            fields=["value", "valueFrom"],
        )
        if envvar.valueFrom:
            yield from require_exactly_one(
                model=envvar.valueFrom,
                loc=("env", i, "valueFrom"),
                fields=[
                    "configMapKeyRef",
                    "fieldRef",
                    "resourceFieldRef",
                    "secretKeyRef",
                ],
            )

    # field `envFrom`
    for i, env_from in enumerate(node.envFrom or ()):
        yield from require_exactly_one(
            model=env_from,
            loc=("envFrom", i),
            fields=["configMapRef", "secretRef"],
        )

    # field `resources`
    if (
        True
        and node.resources
        and node.resources.limits
        and node.resources.requests
        and (request_cpu := node.resources.requests.cpu)
        and (limit_cpu := node.resources.limits.cpu)
        and request_cpu > limit_cpu
    ):
        yield {
            "code": "TPL304",
            "loc": ("resources", "requests", "cpu"),
            "summary": "CPU request exceeds CPU limit",
            "msg": (
                f"The CPU request ({request_cpu}) must be greater than or equal to "
                f"the CPU limit ({limit_cpu}) in template '{template.name}'."
            ),
            "input": str(request_cpu),
        }

    if (
        True
        and node.resources
        and node.resources.limits
        and node.resources.requests
        and (request_memory := node.resources.requests.memory)
        and (limit_memory := node.resources.limits.memory)
        and request_memory > limit_memory
    ):
        yield {
            "code": "TPL304",
            "loc": ("resources", "requests", "memory"),
            "summary": "Memory request exceeds memory limit",
            "msg": (
                f"The memory request ({request_memory}) must be greater than or equal to "
                f"the memory limit ({limit_memory}) in template '{template.name}'."
            ),
            "input": str(request_memory),
        }

    # check model fields references
    for diag in check_model_fields_references(node, context.parameters):
        match diag["code"]:
            case "VAR002":
                if metadata := diag.get("ctx", {}).get("reference"):
                    ref = metadata["found:str"]
                    diag["msg"] = (
                        f"The parameter reference '{ref}' used in template '{template.name}' is invalid."
                    )
        yield diag



from __future__ import annotations

import itertools
import logging
import typing

from pydantic import ValidationError

from tugboat.core import get_plugin_manager
from tugboat.schemas import Manifest
from tugboat.utils import bulk_translate_pydantic_errors

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from tugboat.types import Diagnosis


logger = logging.getLogger(__name__)


def analyze_manifest(manifest: dict) -> list[Diagnosis]:
    """
    Analyze the manifest and return a list of diagnoses.

    This function uses the plugin system to analyze the manifest. The diagnoses
    are collected and returned as a list.

    Parameters
    ----------
    manifest : dict
        The manifest to analyze.

    Returns
    -------
    list[Diagnosis]
        The diagnoses reported by the analyzers.
    """
    # get manifest metadata
    try:
        metadata = get_manifest_metadata(manifest)
    except ValueError as e:
        return [
            {
                "type": "warning",
                "code": "M001",
                "loc": (),
                "summary": "Not a Kubernetes manifest",
                "msg": f"{e}. The input does not look like a Kubernetes manifest.",
            }
        ]

    logger.debug(
        "Starting analysis of manifest %s.%s/%s",
        metadata["group"],
        metadata["kind"],
        metadata.get("name", "<unnamed>"),
    )

    # parse the manifest
    pm = get_plugin_manager()

    try:
        manifest_obj = pm.hook.parse_manifest(manifest=manifest)
    except ValidationError as e:
        diagnoses = bulk_translate_pydantic_errors(e.errors())
        for diagnosis in diagnoses:
            context = diagnosis.setdefault("ctx", {})
            context["manifest"] = metadata
        return diagnoses
    except Exception as e:
        logger.exception("Error during execution of parse_manifest hook")
        return [
            {
                "type": "error",
                "code": "F001",
                "loc": (),
                "summary": "Internal error while analyzing manifest",
                "msg": f"An error occurred while parsing the manifest: {e}",
                "ctx": {"manifest": metadata},
            }
        ]

    logger.debug(
        "Parsed manifest '%s' as %s object",
        metadata.get("name", "<unknown>"),
        type(manifest_obj),
    )

    if not manifest_obj:
        return [
            {
                "type": "warning",
                "code": "M002",
                "loc": ("kind",),
                "summary": "Unsupported manifest kind",
                "msg": f"Manifest kind {metadata['kind']} is not supported",
                "input": metadata["kind"],
                "ctx": {"manifest": metadata},
            }
        ]

    if not isinstance(manifest_obj, Manifest):
        return [
            {
                "type": "error",
                "code": "F001",
                "loc": (),
                "summary": "Internal error while analyzing manifest",
                "msg": f"Expected a Manifest object, got {type(manifest_obj)}",
                "ctx": {"manifest": metadata},
            }
        ]

    # analyze the manifest
    analyzers_diagnoses: Iterable[Iterator[Diagnosis]]
    analyzers_diagnoses = pm.hook.analyze(manifest=manifest_obj)

    try:  # force evaluation
        diagnoses = list(itertools.chain.from_iterable(analyzers_diagnoses))
    except Exception as e:
        logger.exception("Error during execution of analyze hook")
        return [
            {
                "type": "error",
                "code": "F001",
                "loc": (),
                "summary": "Internal error while analyzing manifest",
                "msg": f"An error occurred while analyzing the manifest: {e}",
                "ctx": {"manifest": metadata},
            }
        ]

    logger.debug(
        "Got %d diagnoses for manifest '%s'",
        len(diagnoses),
        metadata.get("name", "<unknown>"),
    )

    # sort the diagnoses
    def _sort_key(diagnosis: Diagnosis) -> tuple:
        return (
            # 1. position of the occurrence
            diagnosis["loc"],
            # 2. diagnosis code
            diagnosis["code"],
        )

    diagnoses = sorted(diagnoses, key=_sort_key)

    # add manifest info to each diagnosis
    for diagnosis in diagnoses:
        context = diagnosis.setdefault("ctx", {})
        context["manifest"] = metadata

    return diagnoses


def get_manifest_metadata(manifest: dict) -> dict[str, str]:
    """Safely extract metadata from the manifest."""
    output: dict[str, str] = {}

    # API group
    api_version = manifest.get("apiVersion")
    if api_version and isinstance(api_version, str):
        if api_version == "v1":
            output["group"] = ""  # core group
        elif "/" in api_version:
            group, _ = api_version.split("/", 1)
            output["group"] = group
        else:
            raise ValueError("Invalid apiVersion format")
    else:
        raise ValueError("Missing apiVersion")

    # kind
    kind = manifest.get("kind")
    if kind and isinstance(kind, str):
        output["kind"] = kind
    else:
        raise ValueError("kind is missing")

    # name
    metadata = manifest.get("metadata")
    if metadata and isinstance(metadata, dict):
        name = metadata.get("name") or metadata.get("generateName")
        if name and isinstance(name, str):
            output["name"] = name

    return output



__all__ = [
    "bulk_translate_pydantic_errors",
    "check_model_fields_references",
    "check_value_references",
    "find_duplicate_names",
    "get_alias",
    "get_context_name",
    "join_with_and",
    "join_with_or",
    "prepend_loc",
    "translate_pydantic_error",
]

from tugboat.utils.humanize import (
    get_alias,
    get_context_name,
    join_with_and,
    join_with_or,
)
from tugboat.utils.operator import (
    check_model_fields_references,
    check_value_references,
    find_duplicate_names,
    prepend_loc,
)
from tugboat.utils.pydantic import (
    bulk_translate_pydantic_errors,
    translate_pydantic_error,
)



"""
:py:mod:`tugboat.hookspecs` manages the hook specifications for Tugboat.
It is used to define the hooks that can be implemented by plugins.

The hooks are defined as functions with specific signatures.
Tugboat calls these hooks during the linting process.

The Tugboat framework heavily relies on :py:mod:`pluggy` for hook management.
For details on how to implement hooks, please refer to the
`pluggy's documentation <https://pluggy.readthedocs.io/en/stable/>`_.
"""



from __future__ import annotations

import typing

from tugboat._vendor.pygments.lexer import RegexLexer, words
from tugboat._vendor.pygments.token import Name, Other, Punctuation, Text, Whitespace

if typing.TYPE_CHECKING:
    from typing import ClassVar


class ArgoTemplateLexer(RegexLexer):
    """
    Argo template language lexer

    See also
    --------
    Template Tag Kinds
       https://argo-workflows.readthedocs.io/en/latest/variables/#template-tag-kinds
    """

    name: ClassVar = "Argo template"

    tokens: ClassVar = {
        "root": [
            (words(("{{=",)), Punctuation, "expression"),
            (words(("{{",)), Punctuation, "simple-reference"),
            (r"[^{]+", Text),
            (r"{", Text),
        ],
        "simple-reference": [
            (r"\s+", Whitespace),
            (r"\.", Punctuation),
            (r"[\w-]+", Name.Variable),
            (r"}}", Punctuation, "#pop"),
        ],
        "expression": [
            (r"[^}]+", Other),
            (r"}}", Punctuation, "#pop"),
        ],
    }



"""
This module defines the hook specifications for Argo's `workflow`_ and
`workflow template`_ analysis.

.. _workflow: https://argo-workflows.readthedocs.io/en/latest/workflow-concepts/#the-workflow
.. _workflow template: https://argo-workflows.readthedocs.io/en/latest/workflow-templates/
"""

from __future__ import annotations

import typing

from tugboat.hookspecs.core import hookspec

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    from tugboat.schemas import Step, Template, Workflow, WorkflowTemplate
    from tugboat.types import Diagnosis


@hookspec
def analyze_workflow(workflow: Workflow) -> Iterable[Diagnosis]:  # type: ignore[reportReturnType]
    """
    Analyze a workflow.

    Parameters
    ----------
    workflow : Workflow
        The workflow to analyze.
    """


@hookspec
def analyze_workflow_template(workflow_template: WorkflowTemplate) -> Iterable[Diagnosis]:  # type: ignore[reportReturnType]
    """
    Analyze a workflow template.

    Parameters
    ----------
    workflow_template : WorkflowTemplate
        The workflow template to analyze.
    """


@hookspec
def analyze_template(template: Template, workflow: Workflow | WorkflowTemplate) -> Iterable[Diagnosis]:  # type: ignore[reportReturnType]
    """
    Analyze a template.

    This hook is called for each template in a workflow or workflow template,
    but not including the :py:attr:`~tugboat.schemas.workflow.WorkflowSpec.templateDefaults`.

    For issues reported by this hook, the ``loc`` attribute of the diagnosis
    should start from the ``template`` object itself. The framework will manage
    the outer structure and its location path.

    Parameters
    ----------
    template : Template
        The template to analyze.
    workflow : Workflow | WorkflowTemplate
        The workflow that the template is part of.
    """


@hookspec
def analyze_step(
    step: Step, template: Template, workflow: Workflow | WorkflowTemplate
) -> Iterable[Diagnosis]:  # type: ignore[reportReturnType]
    """
    Analyze a workflow step.

    For issues reported by this hook, the ``loc`` attribute of the diagnosis
    should start from the ``step`` object itself. The framework will manage
    the outer structure and its location path.

    Parameters
    ----------
    step : Step
        The step to analyze.
    template : Template
        The template that the step is part of.
    workflow : Workflow | WorkflowTemplate
        The workflow that the template is part of.
    """



from __future__ import annotations

import typing

from tugboat.constraints import (
    accept_none,
    mutually_exclusive,
    require_exactly_one,
    require_non_empty,
)
from tugboat.core import hookimpl
from tugboat.references import get_template_context
from tugboat.utils import (
    check_model_fields_references,
    check_value_references,
    find_duplicate_names,
    prepend_loc,
)

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    from tugboat.references import Context
    from tugboat.schemas import (
        Artifact,
        Parameter,
        Template,
        Workflow,
        WorkflowTemplate,
    )
    from tugboat.types import Diagnosis


@hookimpl(specname="analyze_template")
def check_output_parameters(
    template: Template, workflow: Workflow | WorkflowTemplate
) -> Iterable[Diagnosis]:
    if not template.outputs:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(template.outputs.parameters or ()):
        yield {
            "code": "TPL104",
            "loc": ("outputs", "parameters", idx, "name"),
            "summary": "Duplicate parameter name",
            "msg": f"Parameter name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each parameter
    context = get_template_context(workflow, template)

    for idx, param in enumerate(template.outputs.parameters or ()):
        yield from prepend_loc(
            ("outputs", "parameters", idx),
            _check_output_parameter(template, param, context),
        )


def _check_output_parameter(
    template: Template, param: Parameter, context: Context
) -> Iterable[Diagnosis]:
    yield from require_non_empty(
        model=param,
        loc=(),
        fields=["name", "valueFrom"],
    )
    yield from accept_none(
        model=param,
        loc=(),
        fields=["value"],
    )

    if param.valueFrom:
        accept_fields = ["configMapKeyRef", "expression", "parameter"]
        match template.type:
            case "container" | "containerSet" | "script":
                accept_fields += ["path"]
            case "resource":
                accept_fields += ["jqFilter", "jsonPath"]
            case "suspend":
                accept_fields += ["supplied"]

        reject_fields = {"event", "jqFilter", "jsonPath", "path", "supplied"}
        reject_fields.difference_update(accept_fields)

        yield from require_exactly_one(
            model=param.valueFrom, loc=("valueFrom",), fields=accept_fields
        )
        yield from accept_none(
            model=param.valueFrom, loc=("valueFrom",), fields=reject_fields
        )

        # TODO check expression

    for diag in check_model_fields_references(param, context.parameters):
        match diag["code"]:
            case "VAR002":
                if metadata := diag.get("ctx", {}).get("reference"):
                    ref = metadata["found:str"]
                    diag["msg"] = (
                        f"The parameter reference '{ref}' used in parameter '{param.name}' is invalid."
                    )
        yield diag


@hookimpl(specname="analyze_template")
def check_output_artifacts(
    template: Template, workflow: Workflow | WorkflowTemplate
) -> Iterable[Diagnosis]:
    if not template.outputs:
        return

    # report duplicate names
    for idx, name in find_duplicate_names(template.outputs.artifacts or ()):
        yield {
            "code": "TPL105",
            "loc": ("outputs", "artifacts", idx, "name"),
            "summary": "Duplicate artifact name",
            "msg": f"Artifact name '{name}' is duplicated.",
            "input": name,
        }

    # check fields for each artifact
    context = get_template_context(workflow, template)

    for idx, artifact in enumerate(template.outputs.artifacts or ()):
        yield from prepend_loc(
            ("outputs", "artifacts", idx),
            _check_output_artifact(template, artifact, context),
        )


def _check_output_artifact(
    template: Template, artifact: Artifact, context: Context
) -> Iterable[Diagnosis]:
    yield from require_non_empty(
        model=artifact,
        loc=(),
        fields=["name"],
    )
    yield from mutually_exclusive(
        model=artifact,
        loc=(),
        fields=[
            "artifactory",
            "azure",
            "gcs",
            "hdfs",
            "http",
            "oss",
            "s3",
        ],
    )

    accept_source_fields = ["from_", "fromExpression"]
    reject_source_fields = ["git", "raw"]

    if template.type in ("container", "containerSet", "data", "script"):
        accept_source_fields.append("path")
    else:
        reject_source_fields.append("path")

    yield from require_exactly_one(model=artifact, loc=(), fields=accept_source_fields)
    yield from accept_none(model=artifact, loc=(), fields=reject_source_fields)

    if artifact.archive:
        yield from require_exactly_one(
            model=artifact.archive,
            loc=("archive",),
            fields=[
                "none",
                "tar",
                "zip",
            ],
        )

    if artifact.from_:
        for diag in prepend_loc(
            ("from",), check_value_references(artifact.from_, context.artifacts)
        ):
            match diag["code"]:
                case "VAR002":
                    if metadata := diag.get("ctx", {}).get("reference"):
                        ref = metadata["found:str"]
                        diag["msg"] = (
                            f"The artifact reference '{ref}' used in artifact '{artifact.name}' is invalid."
                        )
            yield diag

    # TODO artifact.fromExpression



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class VolumeMount(BaseModel):
    """
    `VolumeMount`_ describes a mounting of a `Volume`_ within a container.

    .. _VolumeMount: https://argo-workflows.readthedocs.io/en/latest/fields/#volumemount
    .. _Volume: https://kubernetes.io/docs/concepts/storage/volumes/
    """

    model_config = ConfigDict(extra="forbid", frozen=True)

    mountPath: str
    mountPropagation: str | None = None
    name: str
    readOnly: bool | None = None
    recursiveReadOnly: bool | None = None
    subPath: str | None = None
    subPathExpr: str | None = None



from __future__ import annotations

import itertools
import logging
import typing

from tugboat.references.cache import cache
from tugboat.references.context import AnyStr
from tugboat.references.workflow import get_workflow_context

if typing.TYPE_CHECKING:
    from tugboat.references.context import Context
    from tugboat.schemas import Template, Workflow, WorkflowTemplate

logger = logging.getLogger(__name__)


@cache(16)
def get_template_context(
    workflow: Workflow | WorkflowTemplate, template: Template
) -> Context:
    """
    Returns a context with the available references for the given template.
    """
    ctx = get_workflow_context(workflow)

    # all template
    ctx.parameters |= {
        ("inputs", "parameters"),
        ("node", "name"),
        ("workflow", "status"),
        ("workflow", "failures"),
    }

    if template.inputs:
        ctx.parameters |= {
            ("inputs", "parameters", param.name)
            for param in template.inputs.parameters or []
        }
        ctx.artifacts |= {
            ("inputs", "artifacts", artifact.name)
            for artifact in template.inputs.artifacts or []
        }

    # container/script template
    if template.container or template.script or template.containerSet:
        ctx.parameters |= {("pod", "name")}

        if template.retryStrategy:
            ctx.parameters |= {("retries",)}

        if template.inputs:
            ctx.parameters |= {
                ("inputs", "artifacts", artifact.name, "path")
                for artifact in template.inputs.artifacts or []
            }

        if template.outputs:
            ctx.parameters |= {
                ("outputs", "artifacts", artifact.name, "path")
                for artifact in template.outputs.artifacts or []
            }
            ctx.parameters |= {
                ("outputs", "parameters", param.name, "path")
                for param in template.outputs.parameters or []
            }

    # add template-type specific references
    _add_step_references(ctx, template, workflow)
    _add_dag_references(ctx, template, workflow)
    # TODO http template

    return ctx


def _add_step_references(
    ctx: Context,
    template: Template,
    workflow: Workflow | WorkflowTemplate,
):
    if not template.steps:
        return

    ctx.parameters |= {("steps", "name")}

    for step in itertools.chain.from_iterable(template.steps):
        # default step parameters
        ctx.parameters |= {
            ("steps", step.name, "id"),
            ("steps", step.name, "ip"),
            ("steps", step.name, "status"),
            ("steps", step.name, "exitCode"),
            ("steps", step.name, "startedAt"),
            ("steps", step.name, "finishedAt"),
            ("steps", step.name, "hostNodeName"),
            ("steps", step.name, "outputs", "result"),
        }

        # parallel steps
        if step.withItems or step.withParam:
            ctx.parameters |= {
                ("steps", step.name, "outputs", "parameters"),
            }

        # step outputs
        # when the referenced template is in the same workflow, we can resolve the outputs
        # otherwise, we can only assume the outputs are available
        reference_template = None
        if step.template:
            reference_template = workflow.template_dict.get(step.template)
        elif step.templateRef and step.templateRef.name == workflow.name:
            reference_template = workflow.template_dict.get(step.templateRef.template)
        elif step.inline:
            reference_template = step.inline

        if reference_template and reference_template.outputs:
            ctx.parameters |= {
                ("steps", step.name, "outputs", "parameters", param.name)
                for param in reference_template.outputs.parameters or []
                if param.name
            }
            ctx.artifacts |= {
                ("steps", step.name, "outputs", "artifacts", artifact.name)
                for artifact in reference_template.outputs.artifacts or []
                if artifact.name
            }

        if not reference_template:
            logger.debug(
                "The referenced template %s is not available. "
                "Allow `steps.%s.outputs.parameters.ANY` and `steps.%s.outputs.artifacts.ANY`.",
                step.template,
                step.name,
                step.name,
            )
            ctx.parameters |= {
                ("steps", step.name, "outputs", "parameters", AnyStr),
            }
            ctx.artifacts |= {
                ("steps", step.name, "outputs", "artifacts", AnyStr),
            }


def _add_dag_references(
    ctx: Context,
    template: Template,
    workflow: Workflow | WorkflowTemplate,
):
    if not template.dag:
        return

    ctx.parameters |= {("tasks", "name")}

    for task in template.dag.tasks:
        # default task parameters
        ctx.parameters |= {
            ("tasks", task.name, "id"),
            ("tasks", task.name, "ip"),
            ("tasks", task.name, "status"),
            ("tasks", task.name, "exitCode"),
            ("tasks", task.name, "startedAt"),
            ("tasks", task.name, "finishedAt"),
            ("tasks", task.name, "hostNodeName"),
            ("tasks", task.name, "outputs", "result"),
        }

        # parallel tasks
        if task.withItems or task.withParam:
            ctx.parameters |= {
                ("tasks", task.name, "outputs", "parameters"),
            }

        # task outputs
        referenced_template = None
        if task.template:
            referenced_template = workflow.template_dict.get(task.template)
        elif task.templateRef and task.templateRef.name == workflow.name:
            referenced_template = workflow.template_dict.get(task.templateRef.template)
        elif task.inline:
            referenced_template = task.inline

        if referenced_template and referenced_template.outputs:
            ctx.parameters |= {
                ("tasks", task.name, "outputs", "parameters", param.name)
                for param in referenced_template.outputs.parameters or ()
                if param.name
            }
            ctx.artifacts |= {
                ("tasks", task.name, "outputs", "artifacts", artifact.name)
                for artifact in referenced_template.outputs.artifacts or []
                if artifact.name
            }

        if not referenced_template:
            logger.debug(
                "The referenced template %s is not available. "
                "Allow `tasks.%s.outputs.parameters.ANY` and `tasks.%s.outputs.artifacts.ANY`.",
                task.template,
                task.name,
                task.name,
            )
            ctx.parameters |= {
                ("tasks", task.name, "outputs", "parameters", AnyStr),
            }
            ctx.artifacts |= {
                ("tasks", task.name, "outputs", "artifacts", AnyStr),
            }



from __future__ import annotations

import contextlib
import logging
import sys
import typing

import click
import cloup
import colorlog
from pydantic import ValidationError

import tugboat.console.anchor
import tugboat.settings
from tugboat.console.formatters import get_output_formatter
from tugboat.console.glob import gather_paths
from tugboat.console.utils import DiagnosesCounter, Stdin
from tugboat.engine import analyze_yaml_stream
from tugboat.settings import settings
from tugboat.version import __version__

if typing.TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Any, NoReturn, TextIO

    from pydantic import FilePath

logger = logging.getLogger(__name__)


@cloup.command(
    context_settings={
        "help_option_names": ["-h", "--help"],
    }
)
@cloup.argument(
    "manifest",
    nargs=-1,
    help="List of files or directories to check. [default: .]",
)
@cloup.option_group(
    "Input options",
    cloup.option(
        "--exclude",
        multiple=True,
        help=(
            "List of files or directories to exclude from the check. "
            "This option can be used multiple times."
        ),
    ),
    cloup.option(
        "--follow-symlinks/--no-follow-symlinks",
        default=None,
        help="Follow symbolic links when checking directories.",
    ),
)
@cloup.option_group(
    "Output options",
    cloup.option(
        "--color/--no-color",
        default=None,
        help="Colorize the output. [default: auto]",
    ),
    cloup.option(
        "--output-format",
        type=cloup.Choice(["console", "junit", "github"]),
        help="Output serialization format. [default: console]",
    ),
    cloup.option(
        "-o",
        "--output-file",
        type=cloup.file_path(writable=True),
        help="File to write the diagnostic output to.",
    ),
)
@cloup.option_group(
    "Logging options",
    cloup.option(
        "-v",
        "--verbose",
        count=True,
        help="Print more information.",
    ),
)
@cloup.option(
    "--mcp",
    is_flag=True,
    help="Start the MCP server.",
)
@cloup.option(
    "--anchor",
    is_flag=True,
    help="Drop an anchor.",
)
@cloup.version_option(__version__)
def main(
    manifest: Sequence[str],
    exclude: Sequence[str],
    follow_symlinks: bool | None,
    color: bool | None,
    output_format: str | None,
    output_file: Path | None,
    verbose: int,
    mcp: bool,
    anchor: bool,
):
    """
    Linter to streamline your Argo Workflows with precision and confidence.

    This command performs static analysis on Argo Workflow manifests to catch
    common issues and improve the quality of your workflows.

    Examples:

    \b
      # Check all YAML files in the current directory
      tugboat

    \b
      # Check specific files
      tugboat my-workflow-1.yaml my-workflow-2.yaml

    \b
      # Analyze all YAML files in a directory
      tugboat my-workflows/

    \b
      # Read from stdin
      cat my-workflow.yaml | tugboat
    """
    # setup logging
    setup_loggings(verbose)
    logger.debug("Tugboat sets sail!")

    # easter egg: drop an anchor
    if anchor:
        tugboat.console.anchor.print_anchor()

    # update and validate settings
    update_settings(**locals())

    # special case: MCP
    if mcp:
        run_mcp()

    # perform linting
    if output_file:
        output_stream = output_file.open("w")
    else:
        # `nullcontext` prevents __exit__ call
        output_stream = typing.cast("TextIO", contextlib.nullcontext(sys.stdout))

    lint(output_stream)


def setup_loggings(verbose_level: int):
    """
    Setup loggings

    We separate the configuration of the tugboat logger from the other loggers.
    """
    # tugboat loggers
    tugboat_logger = colorlog.getLogger("tugboat")
    tugboat_logger.propagate = False

    match verbose_level:
        case 0:
            tugboat_logger.setLevel(colorlog.WARNING)
        case 1:
            tugboat_logger.setLevel(colorlog.INFO)
        case _:
            tugboat_logger.setLevel(colorlog.DEBUG)

    handler = colorlog.StreamHandler()
    handler.setFormatter(
        colorlog.ColoredFormatter("%(log_color)s%(levelname)7s |%(reset)s %(message)s")
    )
    tugboat_logger.addHandler(handler)

    # other loggers
    # only show logs when verbose level >= 3
    if verbose_level >= 3:
        logger = colorlog.getLogger()
        logger.setLevel(colorlog.DEBUG)
        logger.addHandler(handler)


def update_settings(
    manifest: Sequence[str],
    exclude: Sequence[str],
    follow_symlinks: bool | None,
    color: bool | None,
    output_format: str | None,
    **sink,
):
    """
    Update settings using command line arguments.

    This function replaces any settings loaded from config files or environment
    variables with values provided through command line options.

    The updates are applied directly to the settings object.
    """
    # pass the input options to the settings data model for validation
    update_args: dict[str, Any] = {
        "include": manifest,
        "exclude": exclude,
    }

    if color is not None:
        update_args["color"] = color
    if follow_symlinks is not None:
        update_args["follow_symlinks"] = follow_symlinks
    if output_format is not None:
        update_args["output_format"] = output_format

    # update settings
    with reraise_usage_error():
        # https://docs.pydantic.dev/latest/concepts/pydantic_settings/#in-place-reloading
        tugboat.settings.settings.__init__(**update_args)

    # special case: stdin
    if manifest == () and not sys.stdin.isatty() and not sys.stdin.closed:
        logger.debug("Detected stdin. Using it as input.")
        path = typing.cast("FilePath", Stdin())
        tugboat.settings.settings.include = [path]

    # log current settings
    logger.debug(
        "Current settings: %s",
        tugboat.settings.settings.model_dump_json(indent=2),
    )


@contextlib.contextmanager
def reraise_usage_error():
    try:
        yield
    except ValidationError as e:
        # only use the first error
        err, *_ = e.errors()

        match setting_name := err["loc"][0]:
            case "include":
                option_name = "manifest"
            case "exclude":
                option_name = "--exclude"
            case "color":
                option_name = "--color"
            case "follow_symlinks":
                option_name = "--follow-symlinks"
            case "output_format":
                option_name = "--output-format"
            case _:
                option_name = setting_name  # pragma: no cover

        value = err["input"]
        message = err["msg"]
        raise click.UsageError(
            f"Invalid value for '{option_name}': {message} (input= {value})"
        ) from None


def lint(output_stream: TextIO) -> NoReturn:
    # determine the inputs
    manifest_paths = gather_paths(
        includes=settings.include,
        excludes=settings.exclude,
        follow_symlinks=settings.follow_symlinks,
    )

    if not manifest_paths:
        raise click.UsageError("No manifest found.")

    manifest_paths = sorted(manifest_paths)
    logger.info("Found %d manifest(s) to analyze.", len(manifest_paths))

    # analyze manifests
    counter = DiagnosesCounter()
    output_formatter = get_output_formatter()

    for i, path in enumerate(manifest_paths, 1):
        logger.info("[%d/%d] Analyzing file %s", i, len(manifest_paths), path)
        try:
            content = path.read_text()
        except Exception:
            logger.error("Failed to read file %s", path)
            logger.debug("Error details:", exc_info=True)
            raise click.Abort from None

        diagnoses = analyze_yaml_stream(content, path)

        counter.update(diag.type for diag in diagnoses)
        output_formatter.update(content=content, diagnoses=diagnoses)

    logger.debug("Analysis completed. Found %d diagnoses.", sum(counter.values()))

    # write report
    with output_stream as stream:
        output_formatter.dump(stream)

    # finalize
    click.echo(counter.summary(), err=True)

    if counter.has_any_error():
        sys.exit(2)

    sys.exit(0)


def run_mcp() -> NoReturn:
    try:
        import tugboat.console.mcp
    except ImportError:
        raise click.UsageError(
            "MCP is not installed. To use this feature, please install Tugboat with the 'mcp' extra."
        ) from None

    tugboat.console.mcp.server.run()
    sys.exit(0)



from __future__ import annotations

import asyncio
import logging
import textwrap
import typing
from pathlib import Path
from typing import Annotated

from mcp.server.fastmcp import FastMCP
from mcp.types import ToolAnnotations
from pydantic import BaseModel, Field

from tugboat.engine import DiagnosisModel, analyze_yaml_stream

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    from pydantic.fields import FieldInfo

logger = logging.getLogger(__name__)
server = FastMCP("tugboat")


def _Docstring(description: str) -> FieldInfo:
    """Shortcut for creating a description-only Field annotation."""
    description = textwrap.dedent(description).strip()
    return Field(description=description)


class SuccessResult(BaseModel):
    """A report generated by the analyzer."""

    count: Annotated[
        int,
        _Docstring("The number of issues found in the manifest."),
    ]

    issues: Annotated[
        list[Issue],
        _Docstring("A list of issues found in the manifest."),
    ]


class ErrorResult(BaseModel):
    """An error result returned by the analyzer."""

    message: Annotated[
        str,
        _Docstring("A human-readable message describing the error."),
    ]


class Issue(DiagnosisModel):
    """An issue reported by the analyzer."""

    sourceNearby: Annotated[
        str,
        _Docstring("Text near the issue occurrence in the source file."),
    ]


@server.tool(
    annotations=ToolAnnotations(
        title="Analyze Manifest",
        readOnlyHint=True,
        destructiveHint=False,
        openWorldHint=True,
    )
)
async def analyze_stream(
    manifest_path: Annotated[
        str,
        _Docstring(
            """
            Absolute or relative path to the Argo Workflows manifest on disk.
            Provide plain YAML; if you only have a template, render it first or set `is_helm_template`.
            """
        ),
    ],
    is_helm_template: Annotated[
        bool,
        _Docstring(
            """
            Toggle Helm template rendering.
            Set to `true` when the path points to an unrendered Helm template so the tool renders before linting.
            """
        ),
    ] = False,
) -> SuccessResult | ErrorResult:
    """
    Run lint checks on an Argo Workflows manifest and surface any rule violations.

    Each reported issue includes the rule code, location, summary, suggested fix, and nearby source
    context so you can update the manifest or produce a diff.

    ### Example

    Manifest at `/path/to/whalesay.yaml`:

    ```yaml
    apiVersion: argoproj.io/v1alpha1
    kind: Workflow
    metadata:
      generateName: demo-
    spec:
      entrypoint: ducksay
      templates:
        - name: whalesay
          container:
            image: docker/whalesay:latest
    ```

    Analyzer output:

    ```json
    {"count":1,"issues":[{"code":"WF201","summary":"Invalid entrypoint","msg":"Entrypoint 'ducksay' is not defined in any template.","fix":"whalesay"}]}
    ```

    Suggested remediation:

    - Update `spec.entrypoint` to reference an existing template (here: `whalesay`).
    - Example patch:

      ```patch
      @@
      -  entrypoint: ducksay
      +  entrypoint: whalesay
      ```
    """
    # resolve the manifest path
    manifest = Path(manifest_path).resolve()
    logger.debug("Linting manifest %s", manifest)

    if not manifest.exists():
        return ErrorResult.model_validate(
            {
                "message": f"Manifest not found. Input path: {manifest}, resolved path: {manifest}",
            }
        )

    if not manifest.is_file():
        return ErrorResult.model_validate(
            {
                "message": f"Manifest path is not a file. Input path: {manifest}, resolved path: {manifest}",
            }
        )

    # read the manifest content
    if is_helm_template:
        # render Helm template
        try:
            manifest_content = await render_helm_template(manifest)
        except Exception as e:
            return ErrorResult.model_validate({"message": str(e)})

    else:
        # plain manifest file
        with manifest.open() as fd:
            manifest_content = fd.read()

    manifest_content_lines = manifest_content.splitlines()

    # analyze
    issues = []
    for diagnosis in analyze_yaml_stream(manifest_content, manifest):
        issue = diagnosis.model_dump()

        # get lines near the issue
        issue["sourceNearby"] = "\n".join(
            get_lines_near(manifest_content_lines, diagnosis.line)
        )

        issues.append(issue)

    return SuccessResult.model_validate(
        {
            "count": len(issues),
            "issues": issues,
        }
    )


async def render_helm_template(template: Path) -> str:
    """
    Give a path to the template file.
    Returns the rendered output of a helm template command.
    """
    # locate the Helm chart root (directory containing Chart.yaml).
    chart_dir: Path | None = None
    for candidate in template.parents:
        if (candidate / "Chart.yaml").is_file():
            chart_dir = candidate
            break

    if chart_dir is None:
        raise FileNotFoundError(
            f"Could not find Chart.yaml for template path {template}"
        )

    # render the template using `helm template`
    command = [
        "helm",
        "template",
        ".",
        "--show-only",
        str(template.relative_to(chart_dir)),
    ]

    logger.debug("Executing command: $ %s", " ".join(command))

    try:
        process = await asyncio.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=chart_dir,
        )
    except FileNotFoundError:
        raise RuntimeError(
            "Helm executable not found; ensure Helm is installed and in PATH."
        ) from None

    stdout, stderr = await process.communicate()

    if stderr:
        error_msg = stderr.decode("utf-8", errors="replace").strip()
        logger.warning("Helm template command stderr: %s", error_msg)

        if process.returncode != 0:
            raise RuntimeError(f"Helm template command failed: {error_msg}")

    return stdout.decode("utf-8", errors="replace")


def get_lines_near(content: list[str], focus_line: int) -> Iterator[str]:
    focus_line -= 1  # 1-based to 0-based
    line_starting = max(0, focus_line - 1)
    line_ending = min(len(content), focus_line + 2)
    yield from content[line_starting : line_ending + 1]



from __future__ import annotations

import typing

from tugboat.references.cache import cache
from tugboat.references.context import Context

if typing.TYPE_CHECKING:
    from tugboat.schemas import Workflow, WorkflowTemplate


@cache(1)
def get_global_context() -> Context:
    """
    Returns a context with the available global references.
    """
    ctx = Context()
    ctx.parameters |= {
        ("workflow", "name"),
        ("workflow", "namespace"),
        ("workflow", "mainEntrypoint"),
        ("workflow", "serviceAccountName"),
        ("workflow", "uid"),
        ("workflow", "parameters", "json"),
        ("workflow", "annotations", "json"),
        ("workflow", "labels", "json"),
        ("workflow", "creationTimestamp"),
        ("workflow", "creationTimestamp", "RFC3339"),
        ("workflow", "priority"),
        ("workflow", "duration"),
        ("workflow", "scheduledTime"),
        ("workflow", "duration"),
    }
    return ctx


@cache(4)
def get_workflow_context(
    workflow: Workflow | WorkflowTemplate,
) -> Context:
    """
    Returns a context with the available references for the given workflow.
    """
    ctx = get_global_context()

    if workflow.spec.arguments:
        ctx.parameters |= {
            ("workflow", "parameters", param.name)
            for param in workflow.spec.arguments.parameters or []
        }

    for template in workflow.spec.templates or []:
        if template.outputs:
            ctx.parameters |= {
                ("workflow", "outputs", "parameters", param.globalName)
                for param in template.outputs.parameters or []
                if param.globalName
            }
            ctx.artifacts |= {
                ("workflow", "outputs", "artifacts", artifact.globalName)
                for artifact in template.outputs.artifacts or []
                if artifact.globalName
            }

    return ctx



from __future__ import annotations

import itertools
import os
import typing
from pathlib import Path
from typing import Literal

import pydantic_settings
from pydantic import (
    DirectoryPath,
    Field,
    FilePath,
    ValidationError,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict

from tugboat.types import GlobPath

if typing.TYPE_CHECKING:
    from typing import Any

    from pydantic_settings import PydanticBaseSettingsSource


type _PathSpec = FilePath | DirectoryPath | GlobPath


class ConsoleOutputSettings(BaseSettings):
    """Settings for the console output."""

    snippet_lines_ahead: int = 2
    """Number of lines to show before the line with the issue."""

    snippet_lines_behind: int = 2
    """Number of lines to show after the line with the issue."""


class Settings(BaseSettings):

    model_config = SettingsConfigDict(
        env_nested_delimiter="__",
        env_prefix="tugboat_",
        extra="ignore",
        nested_model_default_partial_update=True,
        pyproject_toml_table_header=("tool", "tugboat"),
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            pydantic_settings.TomlConfigSettingsSource(
                settings_cls, _find_config(".tugboat.toml")
            ),
            pydantic_settings.PyprojectTomlConfigSettingsSource(
                settings_cls, _find_config("pyproject.toml")
            ),
        )

    color: bool | None = None
    """Colorize the output. If None, the output is colorized if the output is a terminal."""

    console_output: ConsoleOutputSettings = Field(default_factory=ConsoleOutputSettings)
    """Settings for the console output."""

    exclude: list[_PathSpec] = Field(default_factory=list)
    """List of paths or patterns to exclude from the check."""

    include: list[_PathSpec] = Field(default=[Path.cwd()])
    """List of paths or patterns to include in the check."""

    follow_symlinks: bool = False
    """Follow symbolic links when checking directories."""

    output_format: Literal["console", "junit", "github"] = "console"
    """Output serialization format."""

    @field_validator("color")
    @classmethod
    def _validate_color_(cls, value: bool | None) -> bool | None:
        if os.getenv("FORCE_COLOR"):  # https://force-color.org/
            return True
        if os.getenv("NO_COLOR"):  # https://no-color.org/
            return False
        return value

    @field_validator("include", "exclude", mode="wrap")
    @classmethod
    def _validate_path_(
        cls, value: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ):
        try:
            return handler(value)
        except ValidationError as exc:
            error_items = {}
            for error in exc.errors():
                # the second item is the annotation from type validator
                idx, _ = error["loc"]
                error_items[idx] = error["input"]

            line_errors = []
            for idx, input_ in error_items.items():
                line_errors.append(
                    {
                        "type": "value_error",
                        "loc": (idx,),
                        "input": input_,
                        "ctx": {"error": ValueError("input is not a valid path spec.")},
                    }
                )

            raise ValidationError.from_exception_data(
                typing.cast("str", info.field_name),
                line_errors,
            ) from exc


def _find_config(name: str) -> Path | None:
    cwd = Path.cwd()
    for dir_ in itertools.chain([cwd], cwd.parents):
        path = dir_ / name
        if path.is_file():
            return path


settings = Settings()



from __future__ import annotations

__all__ = ["OutputFormatter", "get_output_formatter"]

import tugboat.settings
from tugboat.console.formatters.base import OutputFormatter


def get_output_formatter() -> OutputFormatter:
    match fmt := tugboat.settings.settings.output_format:
        # fmt: off
        case "console":
            from tugboat.console.formatters.console import ConsoleFormatter
            return ConsoleFormatter()

        case "github":
            from tugboat.console.formatters.github import GitHubFormatter
            return GitHubFormatter()

        case "junit":
            from tugboat.console.formatters.junit import JUnitFormatter
            return JUnitFormatter()

    raise RuntimeError(f"Unsupported output format: {fmt}")



from __future__ import annotations

import json
import logging
import typing

from tugboat.references.cache import cache
from tugboat.references.context import AnyStr
from tugboat.references.template import get_template_context

if typing.TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any

    from tugboat.references.context import Context
    from tugboat.schemas import DagTask, Step, Template, Workflow, WorkflowTemplate

logger = logging.getLogger(__name__)


def get_step_context(
    workflow: Workflow | WorkflowTemplate, template: Template, step: Step
) -> Context:
    """
    Retrieve the context for a specific `step`_.

    .. _step: https://argo-workflows.readthedocs.io/en/latest/walk-through/steps/
    """
    return _get_step_context(workflow, template, step)


def get_task_context(
    workflow: Workflow | WorkflowTemplate, template: Template, task: DagTask
) -> Context:
    """
    Retrieves the context for a `DAG`_ task.

    .. _DAG: https://argo-workflows.readthedocs.io/en/latest/walk-through/dag/
    """
    return _get_step_context(workflow, template, task)


@cache(32)
def _get_step_context(
    workflow: Workflow | WorkflowTemplate, template: Template, step: Step | DagTask
) -> Context:
    ctx = get_template_context(workflow, template)

    if step.withItems:
        ctx.parameters |= {("item",)}
        ctx.parameters |= {
            ("item", field) for field in _collect_item_fields(step.withItems)
        }

    if step.withParam:
        ctx.parameters |= {("item",)}

        fields = _collect_param_fields(step.withParam)
        if fields is None:
            logger.debug(
                "Failed to parse withParam in step %s. Allow `item.ANY`.", step.name
            )
            ctx.parameters |= {("item", AnyStr)}
        else:
            ctx.parameters |= {("item", field) for field in fields}

    if step.withSequence:
        ctx.parameters |= {("item",)}

    return ctx


def _collect_item_fields(with_items: Iterable[Any]) -> set[str]:
    fields = set()
    for item in with_items:
        if isinstance(item, dict):
            fields.update(item)
    return fields


def _collect_param_fields(s: str) -> set[str] | None:
    # try to parse the string as json array, or early return
    # it's common to use references in the withParam field
    try:
        data = json.loads(s)
    except Exception:
        return None
    if not isinstance(data, list):
        return None

    # collect all the fields in the json array
    fields = set()
    for item in data:
        if isinstance(item, dict):
            fields.update(item)

    return fields



from __future__ import annotations

import os
import typing

import frozendict
import pydantic_core.core_schema
from pydantic import BaseModel, ConfigDict

if typing.TYPE_CHECKING:
    from typing import Any

    from pydantic import GetCoreSchemaHandler
    from pydantic_core import CoreSchema

if os.getenv("DOCUTILSCONFIG"):
    __all__ = [
        "Array",
        "ConfigKeySelector",
        "Dict",
        "Empty",
        "KeyValuePair",
        "NameValuePair",
        "PodMetadata",
    ]

type Array[T] = tuple[T, ...]
"""Type alias representing an immutable sequence.
"""


class Dict[TK, TV](frozendict.frozendict[TK, TV]):
    """
    A frozen dictionary type that can be used in Pydantic models.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        if type_args := typing.get_args(source_type):
            key_type, value_type = type_args
        else:
            key_type = value_type = typing.Any

        key_schema = handler.generate_schema(key_type)
        value_schema = handler.generate_schema(value_type)

        return pydantic_core.core_schema.no_info_after_validator_function(
            function=frozendict.frozendict,
            schema=pydantic_core.core_schema.dict_schema(
                keys_schema=key_schema,
                values_schema=value_schema,
            ),
        )


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class ConfigKeySelector(_BaseModel):
    """
    Represents a reference to a key within a ConfigMap or Secret.
    This class is utilized by both the `ConfigMapKeySelector`_ and `SecretKeySelector`_ classes.

    .. _ConfigMapKeySelector:
       https://argo-workflows.readthedocs.io/en/latest/fields/#configmapkeyselector
    .. _SecretKeySelector:
       https://argo-workflows.readthedocs.io/en/latest/fields/#secretkeyselector
    """

    key: str
    name: str
    optional: bool | None = None


class Empty(_BaseModel): ...


class KeyValuePair(_BaseModel):
    key: str
    value: str


class NameValuePair(_BaseModel):
    name: str
    value: str


class PodMetadata(_BaseModel):
    annotations: Dict[str, str] | None = None
    labels: Dict[str, str] | None = None



from __future__ import annotations

import io
import os
import random
import sys
import textwrap
import typing
from dataclasses import dataclass, field

import click

from tugboat.version import __version__

if typing.TYPE_CHECKING:
    from typing import NoReturn


def print_anchor() -> NoReturn:
    from tugboat._vendor.lolcat import LolCat, stdoutWin

    # build image
    ANCHOR = r"""
           _
          (_)       ______            __                __
        <--|-->    /_  __/_  ______ _/ /_  ____  ____ _/ /_
       _   |   _    / / / / / / __ `/ __ \/ __ \/ __ `/ __/
      `\__/ \__/`  / / / /_/ / /_/ / /_/ / /_/ / /_/ / /_
        `-. .-'   /_/  \__,_/\__, /_.___/\____/\__,_/\__/  {version}
           '                /____/
    """

    image = textwrap.dedent(ANCHOR).format(version=__version__).strip("\n")

    # setup lolcat
    @dataclass
    class _Option:

        os: int = field(default_factory=lambda: random.randint(0, 256))
        mode: int = field(default_factory=detect_mode)

        animate: bool = False
        force: bool = False
        freq: float = 0.1
        spread: float = 3.0

    options = _Option()

    # print image
    with io.StringIO(image) as input_buffer, io.StringIO() as output_buffer:
        if os.name == "nt":
            output = stdoutWin()
            output.output = output_buffer
        else:
            output = output_buffer

        lolcat = LolCat(mode=options.mode, output=output)

        input_buffer.seek(0)
        lolcat.cat(input_buffer, options)

        click.echo(output_buffer.getvalue(), nl=False)

    sys.exit(0)


def detect_mode():
    from tugboat._vendor.termcolors import ColorSupport, check_supported_colors

    match check_supported_colors():
        case ColorSupport.TRUE_COLOR | ColorSupport.ANSI_256:
            return 256
        case ColorSupport.BASIC:
            return 16

    return 8  # fallback to 8 colors; the lolcat library does not support no colors



from __future__ import annotations

import re
import typing

from ruamel.yaml.comments import CommentedBase, CommentedMap
from ruamel.yaml.tokens import CommentToken

from tugboat.engine.types import DiagnosisModel

if typing.TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Any

    from ruamel.yaml.error import MarkedYAMLError

pattern_noqa_all = re.compile(r"[ ]*#[ ]*noqa(?:;|$)", re.IGNORECASE | re.MULTILINE)
pattern_noqa_line = re.compile(
    r"[ ]*#[ ]*noqa:[ ]*"  # prefix
    r"("
    r"[a-z]+\d+"  # first code
    r"(?:,[ ]*[A-Z]+[0-9]+)*"  # additional codes, separated by commas
    r")",
    re.IGNORECASE,
)


def translate_marked_yaml_error(err: MarkedYAMLError) -> DiagnosisModel:
    """
    Translate a MarkedYAMLError into a more user-friendly format.

    Parameters
    ----------
    err : MarkedYAMLError
        The error to translate.

    Returns
    -------
    DiagnosisModel
        The translated error.
    """
    line = column = 1
    if err.problem_mark:
        line = err.problem_mark.line + 1
        column = err.problem_mark.column + 1

    msg = err.problem or err.context
    if msg and err.context_mark:
        msg += f"\n{err.context_mark}"  # context_mark is not a string

    return DiagnosisModel.model_validate(
        {
            "line": line,
            "column": column,
            "type": "error",
            "code": "F002",
            "loc": (),
            "summary": "Malformed YAML document",
            "msg": msg,
        }
    )


def get_suppression_codes(doc: CommentedMap, loc: Sequence[int | str]) -> Iterator[str]:
    """
    Get suppression codes from noqa comments at the location and all parent locations.

    This checks for noqa comments at the exact location and all parent locations.
    """
    current_node = doc

    # check root level comment first
    yield from _extract_noqa_codes_from_node(current_node, None)

    # navigate through the path, checking for noqa at each level
    for part in loc:
        # check if the current key has a noqa comment
        if isinstance(current_node, CommentedMap):
            yield from _extract_noqa_codes_from_node(current_node, part)

        # navigate to the next level
        try:
            current_node = current_node[part]  # type: ignore[reportIndexIssue]
        except (KeyError, IndexError, TypeError):
            break

        # check if the node itself has a noqa comment
        yield from _extract_noqa_codes_from_node(current_node, None)

    return []


def _extract_noqa_codes_from_node(
    node: CommentedBase, key: int | str | None = None
) -> Iterator[str]:
    """
    Extract series codes from noqa comment on a specific node or key.

    Parameters
    ----------
    node : CommentedBase
        The node to check for comments
    key : int | str | None
        If provided, check for key comment. If None, check node's own comment.

    Returns
    -------
    Iterator[str]
        Iterator of series codes found in the noqa comment.
    """
    if not isinstance(node, CommentedMap):
        return

    # try to find a comment on the key itself
    if key is not None and (key_comment_info := node.ca.items.get(key)):
        _, _, post_value_comment, pre_value_comments = key_comment_info
        if post_value_comment:
            yield from parse_noqa_codes(post_value_comment.value)

        if isinstance(node[key], int | float | str | bool):
            for comment in pre_value_comments or ():
                if isinstance(comment, CommentToken):
                    yield from parse_noqa_codes(comment.value)
                else:
                    yield from parse_noqa_codes(comment)

    # then, try to find an end-of-line comment on the node itself
    if node.ca.comment:
        post_comment, _pre_comments = node.ca.comment
        if post_comment:
            yield from parse_noqa_codes(post_comment.value)


def parse_noqa_codes(text: str) -> Iterator[str]:
    """Parse series codes from a noqa comment."""
    if pattern_noqa_all.match(text):
        yield _Anything("Anything")
        return  # pragma: no cover; the caller may breaks the loop

    if m := pattern_noqa_line.match(text):
        for code in m.group(1).split(","):
            yield code.strip().upper()


class _Anything(str):
    """A string that matches anything."""

    def __eq__(self, other: Any):
        return True



from __future__ import annotations

import re
import typing

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    from tugboat.types import Diagnosis

GENERATED_SUFFIX_LENGTH = 5


def check_resource_name(
    name: str,
    *,
    min_length: int = 1,
    max_length: int = 253,
    is_generate_name: bool = False,
) -> Iterator[Diagnosis]:
    """
    Check if the name is valid to be used as a Kubernetes resource name.

    Yield
    -----
    :rule:`m302` for too short or too long name.
    :rule:`m302` for invalid characters in the name.

    See also
    --------
    `Object Names and IDs
    <https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#names>`_
    """
    if is_generate_name:
        # although kubernetes automatically truncates the name when it exceeds
        # the length, we want to keep the name as close as possible to the
        # original name to help users identify the resource.
        min_length = max(min_length - GENERATED_SUFFIX_LENGTH, 1)
        max_length -= GENERATED_SUFFIX_LENGTH

    if not name:
        yield {
            "type": "failure",
            "code": "M302",
            "loc": (),
            "summary": "Resource name is too short",
            "msg": f"Resource name is empty, minimum length is {min_length}.",
        }
    elif len(name) < min_length:
        yield {
            "type": "failure",
            "code": "M302",
            "loc": (),
            "summary": "Resource name is too short",
            "msg": f"Resource name '{name}' is too short, minimum length is {min_length}.",
            "input": name,
        }

    if len(name) > max_length:
        yield {
            "type": "failure",
            "code": "M302",
            "loc": (),
            "summary": "Resource name is too long",
            "msg": f"Resource name '{name}' is too long, maximum length is {max_length}.",
            "input": name,
        }

    if is_generate_name:
        normalized_name = name.removesuffix("-")
    else:
        normalized_name = name

    pattern = re.compile(r"[a-z0-9]([a-z0-9.-]*[a-z0-9])?")
    if pattern.fullmatch(normalized_name):
        return

    diagnosis: Diagnosis = {
        "type": "failure",
        "code": "M301",
        "loc": (),
        "summary": "Invalid resource name",
        "msg": f"""
            Resource name '{name}' contains invalid characters.
            It must consist of lower case alphanumeric characters, '-' or '.', and must start and end with an alphanumeric character.
            """,
        "input": name,
    }

    alternative_name = normalized_name.replace("_", "-").lower()
    if pattern.fullmatch(alternative_name):
        if is_generate_name and name.endswith("-"):
            alternative_name += "-"
        diagnosis["fix"] = alternative_name

    yield diagnosis



from __future__ import annotations

import typing

from tugboat.analyzers.kubernetes import GENERATED_SUFFIX_LENGTH, check_resource_name
from tugboat.constraints import require_exactly_one
from tugboat.core import hookimpl

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    from tugboat.schemas import CronWorkflow
    from tugboat.types import Diagnosis


@hookimpl
def analyze(manifest: CronWorkflow) -> Iterator[Diagnosis]:
    # early escape if the manifest is not recognized
    if manifest.apiVersion != "argoproj.io/v1alpha1" or manifest.kind != "CronWorkflow":
        return

    # invoke checks
    yield from check_metadata(manifest)


# ----------------------------------------------------------------------------
# CronWorkflow analyzers
# ----------------------------------------------------------------------------

MAX_CRON_WORKFLOW_NAME_LENGTH = 52
"""
The maximum length of a CronWorkflow name.
Ref: https://github.com/argoproj/argo-workflows/blob/v3.5.6/workflow/validate/validate.go#L90-L93
"""


def check_metadata(workflow: CronWorkflow) -> Iterator[Diagnosis]:
    yield from require_exactly_one(
        model=workflow.metadata,
        loc=("metadata",),
        fields=["name", "generateName"],
    )

    if workflow.metadata.name:
        for diagnosis in check_resource_name(
            workflow.metadata.name, max_length=MAX_CRON_WORKFLOW_NAME_LENGTH
        ):
            diagnosis["loc"] = "metadata", "name"

            if diagnosis["code"] == "M302":
                diagnosis["code"] = "CW001"
                diagnosis["msg"] = (
                    f"""
                    The maximum length of a CronWorkflow name is {MAX_CRON_WORKFLOW_NAME_LENGTH} characters.

                    When a Workflow is created, Argo Workflows appends a timestamp to the CronWorkflow name.
                    Make sure the CronWorkflow name is short enough to fit the timestamp, or it may cause creation failures.
                    """
                )

            yield diagnosis

    if workflow.metadata.generateName:
        for diagnosis in check_resource_name(
            workflow.metadata.generateName,
            max_length=MAX_CRON_WORKFLOW_NAME_LENGTH,
            is_generate_name=True,
        ):
            diagnosis["loc"] = "metadata", "generateName"

            if diagnosis["code"] == "M302":
                diagnosis["code"] = "CW001"
                length = MAX_CRON_WORKFLOW_NAME_LENGTH - GENERATED_SUFFIX_LENGTH
                diagnosis["msg"] = (
                    f"""
                    The maximum length of a CronWorkflow generate name is {length} characters.

                    When a Workflow is created, Argo Workflows appends a timestamp to the CronWorkflow name.
                    Make sure the CronWorkflow name is short enough to fit the timestamp, or it may cause creation failures.
                    """
                )

            yield diagnosis



from __future__ import annotations

import typing

from tugboat._vendor.pygments.token import Name, Punctuation, Text
from tugboat.parsers.ast.core import (
    Node,
    Unexpected,
    consume_whitespaces,
    next_lexeme_is,
)

if typing.TYPE_CHECKING:
    from collections.abc import Iterator

    from tugboat.parsers.ast.core import Lexeme

type ReferenceTuple = tuple[str, ...]


class Document(Node):

    @classmethod
    def parse(cls, lexemes: list[Lexeme]) -> Document:
        children = []
        while lexemes:
            for node_cls in [PlainText, SimpleReferenceTag, ExpressionTag, Unexpected]:
                if node := node_cls.parse(lexemes):
                    children.append(node)
                    break
        return cls(children=children)

    def iter_references(self) -> Iterator[tuple[SimpleReferenceTag, ReferenceTuple]]:
        """
        Iterate over all references in this document.
        """
        for child in self.children:
            if isinstance(child, SimpleReferenceTag):
                yield child, child.reference


class PlainText(Node):

    text: str

    @classmethod
    def parse(cls, lexemes: list[Lexeme]) -> PlainText | None:
        buffer = []
        while next_lexeme_is(lexemes, Text):
            _, _, text = lexemes.pop(0)
            buffer.append(text)
        return cls(text="".join(buffer)) if buffer else None


class SimpleReferenceTag(Node):

    raw: str
    """The raw text of the reference tag."""
    reference: ReferenceTuple
    """Reference to a variable."""

    @classmethod
    def parse(cls, lexemes: list[Lexeme]) -> SimpleReferenceTag | Unexpected | None:
        """
        Parse a simple reference tag.

        A typical lexemes sequence is like:

        .. code-block:: python

            [
                (0, Punctuation, "{{"),
                (2, Name.Variable, "foo"),
                (5, Punctuation, "."),
                (6, Name.Variable, "bar"),
                (9, Whitespace, "  "),
                (11, Punctuation, "}}"),
            ]
        """
        # early return
        if not next_lexeme_is(lexemes, token=Punctuation, text="{{"):
            return

        # consume the opening tag and whitespaces
        components = [
            lexemes.pop(0),
            *consume_whitespaces(lexemes),
        ]

        # consume the reference
        reference = []

        if next_lexeme_is(lexemes, Name.Variable):
            _, _, text = lexeme = lexemes.pop(0)
            components.append(lexeme)
            reference.append(text)

            while (
                True
                and next_lexeme_is(lexemes, Punctuation, ".")
                and next_lexeme_is(lexemes, Name.Variable, shift=1)
            ):
                # the dot
                lexeme = lexemes.pop(0)
                components.append(lexeme)

                # the name
                _, _, text = lexeme = lexemes.pop(0)
                components.append(lexeme)
                reference.append(text)

        # consume trailing whitespaces
        components += consume_whitespaces(lexemes)

        # expect the closing tag
        if not next_lexeme_is(lexemes, Punctuation, "}}"):
            if not lexemes or next_lexeme_is(lexemes, Text):
                return Unexpected.create(components, msg="expect closing tag '}}'")
            return Unexpected.create(components)

        components.append(lexemes.pop(0))  # consume the closing tag
        return cls(
            raw="".join(text for _, _, text in components),
            reference=tuple(reference),
        )

    def __str__(self) -> str:
        return self.raw

    @classmethod
    def format(cls, reference: ReferenceTuple) -> str | None:
        """
        Create a string representation of the reference in the same format as
        this reference tag type.
        """
        return "{{ " + ".".join(reference) + " }}" if reference else None


class ExpressionTag(Node):
    """
    This class recognizes the expression tag format and stores the literal content.
    """

    literal: str

    @classmethod
    def parse(cls, lexemes: list[Lexeme]) -> ExpressionTag | Unexpected | None:
        # early return
        if not next_lexeme_is(lexemes, token=Punctuation, text="{{="):
            return

        # consume the opening tag and whitespaces
        components = [
            lexemes.pop(0),
            *consume_whitespaces(lexemes),
        ]

        # consume all lexemes until the closing tag or the end of the lexemes
        while (
            True
            and lexemes
            and not next_lexeme_is(lexemes, Text)
            and not next_lexeme_is(lexemes, Punctuation, "}}")
        ):
            components.append(lexemes.pop(0))

        # expect the closing tag
        if not next_lexeme_is(lexemes, Punctuation, "}}"):
            if not lexemes or next_lexeme_is(lexemes, Text):
                return Unexpected.create(components, msg="expect closing tag '}}'")
            return Unexpected.create(components)

        components.append(lexemes.pop(0))  # consume the closing tag
        return cls(literal="".join(text for _, _, text in components))



"""
This module provides the parser for Argo template syntax.

It parses the input text into an abstract syntax tree (AST) that can be used to
analyze the template and generate a report.
"""

from __future__ import annotations

__all__ = [
    "Document",
    "Node",
    "SimpleReferenceTag",
    "Unexpected",
    "parse_template",
    "report_syntax_errors",
]

import typing

import tugboat.parsers.ast.argo_template
import tugboat.parsers.lexer
from tugboat.parsers.ast import Document, Node, SimpleReferenceTag, Unexpected

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from tugboat.parsers.ast.core import Lexeme
    from tugboat.types import Diagnosis

_argo_template_lexer = None


def parse_template(text: str) -> Document:
    """Parse Argo template text into AST."""
    global _argo_template_lexer
    if not _argo_template_lexer:
        _argo_template_lexer = tugboat.parsers.lexer.ArgoTemplateLexer()

    lexemes: Iterable[Lexeme] = _argo_template_lexer.get_tokens_unprocessed(text)
    return tugboat.parsers.ast.argo_template.Document.parse(list(lexemes))


def report_syntax_errors(node: Node) -> Iterator[Diagnosis]:
    """Report errors in the AST."""
    # direct return if the node is an Unexpected node
    if isinstance(node, Unexpected):
        msg = f"Invalid syntax near '{node.text}'"
        if node.msg:
            msg += f": {node.msg}"
        yield {
            "code": "VAR001",
            "loc": (),
            "summary": "Syntax error",
            "msg": msg,
            "input": node.text,
        }

    # iterate over the children and yield any syntax errors
    i = 0
    while i < len(node.children):
        # try to find consecutive `Unexpected` nodes, if any
        #
        # The AST builder is intend to create a single Unexpected node for each
        # invalid token, so we need to group them together to provide a more
        # readable error message to the user.
        #
        # A special case is when the `Unexpected` node has a message, we yield
        # it as a single diagnosis. This is because the message is likely to be
        # more informative than the concatenated text of the nodes.
        unexpected_nodes = []
        while (
            i < len(node.children)
            and isinstance(child := node.children[i], Unexpected)
            and not child.msg
        ):
            unexpected_nodes.append(child)
            i += 1

        # if we found any, yield a single diagnosis
        if unexpected_nodes:
            text = "".join(node.text for node in unexpected_nodes)
            yield {
                "code": "VAR001",
                "loc": (),
                "summary": "Syntax error",
                "msg": (
                    f"""
                    Failed to parse the Argo workflow variable due to a syntax error.

                    This error is likely caused by an invalid character or an unexpected token in the variable definition.
                    The syntax error is near the following text: {text}
                    """
                ),
                "input": text,
            }
            # TODO
            # ideally we should yield each abnormal node as a separate diagnosis
            # but currently reporting does not look good, so we just return
            # the first one and skip the rest.
            return

        else:
            # recursively check the children
            yield from report_syntax_errors(node.children[i])
            i += 1



from __future__ import annotations

import typing
from abc import ABC, abstractmethod

if typing.TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import TextIO

    from tugboat.engine import DiagnosisModel


class OutputFormatter(ABC):
    """Abstract base class for building output."""

    @abstractmethod
    def update(self, *, content: str, diagnoses: Sequence[DiagnosisModel]) -> None:
        """Update the data to be serialized."""

    @abstractmethod
    def dump(self, stream: TextIO) -> None:
        """Serialize the data to a stream."""



from __future__ import annotations

import collections
import typing

from tugboat.analyzers.metrics import check_prometheus
from tugboat.constraints import require_exactly_one, require_non_empty
from tugboat.core import get_plugin_manager, hookimpl
from tugboat.references import get_template_context
from tugboat.utils import prepend_loc

if typing.TYPE_CHECKING:
    from collections.abc import Iterable

    from tugboat.schemas import Template, Workflow, WorkflowTemplate
    from tugboat.types import Diagnosis

    type WorkflowCompatible = Workflow | WorkflowTemplate


@hookimpl
def analyze_template(template: Template) -> Iterable[Diagnosis]:
    yield from require_non_empty(
        model=template,
        loc=(),
        fields=["name"],
    )
    yield from require_exactly_one(
        model=template,
        loc=(),
        fields=[
            "container",
            "containerSet",
            "dag",
            "data",
            "http",
            "resource",
            "script",
            "steps",
            "suspend",
        ],
    )


@hookimpl(specname="analyze_template")
def check_steps(
    template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    if not template.steps:
        return

    # check for duplicate step names
    step_names = collections.defaultdict(list)
    for idx_stage, stage in enumerate(template.steps or []):
        for idx_step, step in enumerate(stage):
            if step.name:
                step_names[step.name].append(("steps", idx_stage, idx_step, "name"))

    for name, locs in step_names.items():
        if len(locs) > 1:
            for loc in locs:
                yield {
                    "code": "STP101",
                    "loc": loc,
                    "summary": "Duplicate step name",
                    "msg": f"Step name '{name}' is duplicated.",
                    "input": name,
                }

    # check for step definitions
    pm = get_plugin_manager()

    for idx_stage, stage in enumerate(template.steps or ()):
        for idx_step, step in enumerate(stage):
            yield from prepend_loc.from_iterables(
                ["steps", idx_stage, idx_step],
                pm.hook.analyze_step(step=step, template=template, workflow=workflow),
            )


@hookimpl(specname="analyze_template")
def check_metrics(
    template: Template, workflow: WorkflowCompatible
) -> Iterable[Diagnosis]:
    if not template.metrics:
        return

    # build context
    # some additional variables are available when emitting metrics in a template
    # https://argo-workflows.readthedocs.io/en/latest/variables/#metrics
    ctx = get_template_context(workflow, template)
    ctx.parameters |= {
        ("duration",),
        ("exitCode",),
        ("outputs", "result"),
        ("resourcesDuration", "cpu"),
        ("resourcesDuration", "memory"),
        ("retries",),
        ("status",),
    }
    if template.inputs:
        for param in template.inputs.parameters or ():
            ctx.parameters |= {("inputs", "parameters", param.name)}
    if template.outputs:
        for param in template.outputs.parameters or ():
            ctx.parameters |= {("outputs", "parameters", param.name)}

    # check metrics
    for idx, prom in enumerate(template.metrics.prometheus or ()):
        for diagnosis in prepend_loc(
            ("metrics", "prometheus", idx), check_prometheus(prom, ctx)
        ):
            match diagnosis["code"]:
                case "internal:invalid-metric-name":
                    diagnosis["code"] = "TPL301"
                case "internal:invalid-metric-label-name":
                    diagnosis["code"] = "TPL302"
                case "internal:invalid-metric-label-value":
                    diagnosis["code"] = "TPL303"
            yield diagnosis



from __future__ import annotations

import glob
import os
import re
import typing
from pathlib import Path
from typing import TypedDict

import pydantic_core.core_schema

import tugboat._vendor.glob

if typing.TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Any, Literal, NotRequired

    from pydantic import GetCoreSchemaHandler
    from pydantic_core import CoreSchema


class Diagnosis(TypedDict):
    """
    A diagnosis reported by the analyzer.

    This class is a :py:class:`~typing.TypedDict` that defines the structure of
    a diagnosis produced by the analyzer. It serves as the standard format for
    reporting issues detected during analysis.

    All analyzers return diagnoses in this format to ensure consistency across
    different implementations.
    """

    type: NotRequired[Literal["error", "failure", "warning"]]
    """
    The diagnosis type.
    When not provided, it defaults to "failure".

    * ``error`` indicates a critical issue that prevents the analyzer from running.
    * ``failure`` indicates an issue that the analyzer has detected.
    * ``warning`` indicates a potential issue that the analyzer has detected.
      This is not a critical issue, but it may require attention.
    """

    code: str
    """A unique identifier representing the violated rule."""

    loc: Sequence[str | int]
    """
    The location of the issue occurrence within the manifest, specified in a
    path-like format.

    The first element is the key of the manifest, and the rest are the keys of
    the nested dictionaries.
    """

    summary: NotRequired[str]
    """
    The summary.
    When not provided, the first sentence of the message will be used.
    """

    msg: str
    """
    The detailed message.

    When multiple lines are used in the message, the framework will automatically
    dedent it. This allows the analyzer to use Python multiline strings without
    concern for indentation.
    """

    input: NotRequired[Any]
    """
    The input that caused the issue.
    Use :py:class:`Field` to indicate a specific field rather than a value.
    """

    fix: NotRequired[str | None]
    """Possible alternative value to fix the issue."""

    ctx: NotRequired[dict[str, Bundle]]
    """
    Additional context(s).
    The context is a dictionary that maps context names to their corresponding
    bundle.
    """


type Bundle = dict[str, Any]


class PathLike:
    """
    A base class that implements the :py:class:`os.PathLike` interface and works
    with :py:mod:`pydantic` for serialization and deserialization.
    """

    def __init__(self, representation: str):
        self._representation = representation

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:

        def _validator(v: Any):
            if isinstance(v, str):
                v = cls(v)
            return v

        python_schema = pydantic_core.core_schema.no_info_before_validator_function(
            _validator, pydantic_core.core_schema.is_instance_schema(cls)
        )
        json_schema = pydantic_core.core_schema.no_info_after_validator_function(
            str, pydantic_core.core_schema.str_schema()
        )
        return pydantic_core.core_schema.json_or_python_schema(
            python_schema=python_schema,
            json_schema=json_schema,
            serialization=pydantic_core.core_schema.to_string_ser_schema(),
        )

    def __fspath__(self) -> str:
        return self._representation

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._representation!r})"

    def __str__(self) -> str:
        return self._representation


class GlobPath(PathLike):
    """Wraps a glob pattern for path matching."""

    def __init__(self, pattern: str | os.PathLike):
        pattern = str(pattern)
        if "*" not in pattern and "?" not in pattern:
            raise ValueError(f"Pattern '{pattern}' is not a glob pattern")

        pattern = os.path.realpath(pattern)
        super().__init__(pattern)

        regex_pattern = tugboat._vendor.glob.translate(
            pattern, recursive=True, include_hidden=True
        )
        self._regex_pattern = re.compile(regex_pattern)

    def __eq__(self, value) -> bool:
        if isinstance(value, GlobPath):
            value = str(value)
        path = os.path.realpath(value)
        return self._regex_pattern.match(path) is not None

    def iglob(
        self,
        *,
        recursive: bool = False,
        include_hidden: bool = False,
    ) -> Iterator[Path]:
        """Iterate over the files that match the pattern."""
        for item in glob.iglob(
            self._representation,
            recursive=recursive,
            include_hidden=include_hidden,
        ):
            yield Path(item)


class Field(str):
    """
    Representing a field in a YAML document.

    This class is tended to be used as :py:attr:`Diagnosis.input` to indicate
    that the diagnosis is related to a specific field in the YAML document.
    """



from __future__ import annotations

import enum
import io
import textwrap
import typing
from dataclasses import dataclass

import click

from tugboat.console.formatters.base import OutputFormatter
from tugboat.settings import settings

if typing.TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Any, Literal, TextIO

    from tugboat.engine import DiagnosisModel
    from tugboat.settings import ConsoleOutputSettings

    type Color = Literal[
        "black",
        "red",
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
        "white",
    ]


class ConsoleFormatter(OutputFormatter):

    def __init__(self):
        super().__init__()
        self.buf = io.StringIO()

    def update(self, *, content: str, diagnoses: Sequence[DiagnosisModel]) -> None:
        snippet = Snippet(content.splitlines())
        for diagnosis in diagnoses:
            self.buf.write(
                str(
                    DiagnosticMessageBuilder(
                        diagnosis=diagnosis,
                        snippet=snippet,
                        settings=settings.console_output,
                    )
                )
            )
            self.buf.write("\n")

    def dump(self, stream: TextIO) -> None:
        click.echo(
            self.buf.getvalue(),
            file=stream,
            color=settings.color,
            nl=False,
        )


@dataclass
class DiagnosticMessageBuilder:

    diagnosis: DiagnosisModel
    snippet: Snippet
    settings: ConsoleOutputSettings

    @property
    def emphasis(self) -> Style:
        match self.diagnosis.type:
            case "error" | "failure":
                return Style.Error
            case "warning":
                return Style.Warn
        return Style.Error

    def __str__(self) -> str:
        """
        Formats the diagnosis to rich text.

        Example output:

        ```none
        T01 Example error message
          @manifest.yaml:16:11 (demo-)
          @Template:templates/workflow.yaml

          14 |         command: [cowsay]
          15 |         args:
          16 |           - "{{ inputs.parameter.message }}"
             |              ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
             |              └ T01 at .spec.templates[0].container.args[0]

          Detail message explaining the error here.

          Do you mean: {{ inputs.parameters.message }}
        ```
        """
        with io.StringIO() as buf:
            buf.write(self.summary())
            buf.write("\n")
            buf.write(self.code_area())

            if details := self.details():
                buf.write("\n")
                buf.write(details)

            if suggestion := self.suggestion():
                buf.write("\n")
                buf.write(suggestion)

            return buf.getvalue()

    def summary(self) -> str:
        with io.StringIO() as buf:
            # Headline
            # > T01 Example error message
            buf.write(self.emphasis.fmt(self.diagnosis.code))
            buf.write(" ")
            buf.write(Style.Summary.fmt(self.diagnosis.summary))
            buf.write("\n")

            # Path, location, manifest
            # > @manifest.yaml:16:11 (demo-)
            buf.write(Style.PathDelimiter.fmt("  @"))
            if self.diagnosis.extras.file:
                if self.diagnosis.extras.file.is_stdin:
                    buf.write(Style.LocationStdin.fmt("<stdin>"))
                else:
                    buf.write(self.diagnosis.extras.file.filepath)

            buf.write(Style.PathDelimiter.fmt(":"))
            buf.write(str(self.diagnosis.line))
            buf.write(Style.PathDelimiter.fmt(":"))
            buf.write(str(self.diagnosis.column))

            if self.diagnosis.extras.manifest and self.diagnosis.extras.manifest.name:
                buf.write(
                    Style.ManifestName.fmt(f" ({self.diagnosis.extras.manifest.name})")
                )

            buf.write("\n")

            # Helm template info
            # > @Template:templates/workflow.yaml
            if self.diagnosis.extras.helm:
                buf.write(Style.PathDelimiter.fmt("  @Template:"))
                buf.write(self.diagnosis.extras.helm.template)
                buf.write("\n")

            return buf.getvalue()

    def code_area(self) -> str:
        max_line_number = self.diagnosis.line + self.settings.snippet_lines_behind
        lncw = len(str(max_line_number - 1)) + 2  # line number column width

        with io.StringIO() as buf:
            # code snippet before (and including) the issue
            # > 15 |         args:
            # > 16 |           - "{{ inputs.parameter.message }}"
            for line_no, line_text in self.snippet.lines_between(
                start=self.diagnosis.line - self.settings.snippet_lines_ahead,
                last=self.diagnosis.line,
            ):
                buf.write(Style.LineNumber.fmt(f"{line_no:{lncw}} | "))
                buf.write(line_text)
                buf.write("\n")

            # markers for the issue line
            # marker prefix: line number column + " | "
            prefix = Style.LineNumber.fmt(" " * lncw + " | ")

            # draw underline if possible
            # >    ^^^^^^^^^
            if rng := calc_highlight_range(
                self.snippet[self.diagnosis.line],
                offset=self.diagnosis.column - 1,
                substr=self.diagnosis.input,
            ):
                col_start, col_end = rng
                prefix += " " * col_start

                buf.write(prefix)
                buf.write(self.emphasis.fmt("^" * (col_end - col_start)))
                buf.write("\n")

            else:
                prefix += " " * max(self.diagnosis.column - 1, 0)

            # draw position indicator
            # >    └ T01 at .spec.templates[0].container.args[0]
            buf.write(prefix)
            buf.write(self.emphasis.fmt(f"└ {self.diagnosis.code}"))
            buf.write(Style.LocationDelimiter.fmt(" at "))
            buf.write(Style.Location.fmt(self.diagnosis.loc_path))
            buf.write("\n")

            # code snippet after the issue
            for line_no, line_text in self.snippet.lines_between(
                start=self.diagnosis.line + 1,
                last=self.diagnosis.line + self.settings.snippet_lines_behind,
            ):
                buf.write(Style.LineNumber.fmt(f"{line_no:{lncw}} | "))
                buf.write(line_text)
                buf.write("\n")

            return buf.getvalue()

    def details(self) -> str:
        if not self.diagnosis.msg:
            return ""
        return textwrap.indent(self.diagnosis.msg, "  ") + "\n"

    def suggestion(self) -> str:
        if not self.diagnosis.fix:
            return ""

        with io.StringIO() as buf:
            buf.write("  ")
            buf.write(Style.DoYouMean.fmt("Do you mean: "))

            if "\n" in self.diagnosis.fix:
                buf.write("|-\n")
                for line in self.diagnosis.fix.splitlines():
                    buf.write("  ")
                    buf.write(Style.Suggestion.fmt(line))
                    buf.write("\n")
            else:
                buf.write(Style.Suggestion.fmt(self.diagnosis.fix))
                buf.write("\n")

            return buf.getvalue()


class Style(enum.StrEnum):

    fg: Color | None
    bg: Color | None
    bold: bool | None
    dim: bool | None
    underline: bool | None

    def __new__(
        cls,
        value: str,
        fg: Color | None = None,
        bg: Color | None = None,
        bold: bool | None = None,
        dim: bool | None = None,
        underline: bool | None = None,
    ):
        obj = str.__new__(cls)
        obj._value_ = value
        obj.fg = fg
        obj.bg = bg
        obj.bold = bold
        obj.dim = dim
        obj.underline = underline
        return obj

    def fmt(self, text: Any) -> str:
        return click.style(
            str(text),
            fg=self.fg,
            bg=self.bg,
            bold=self.bold,
            dim=self.dim,
            underline=self.underline,
        )

    # fmt: off
    #                   value------- fg------  bg--- bold- dim-- underline
    DoYouMean =         enum.auto(), "cyan"
    Error =             enum.auto(), "red",    None, True
    LineNumber =        enum.auto(), None,     None, None, True
    Location =          enum.auto(), "cyan"
    LocationDelimiter = enum.auto(), None,     None, None, True
    LocationStdin =     enum.auto(), None,     None, True, True
    ManifestName =      enum.auto(), "blue"
    PathDelimiter =     enum.auto(), "cyan"
    Suggestion =        enum.auto(), None,     None, None, None, True
    Summary =           enum.auto(), None,     None, True
    Warn =              enum.auto(), "yellow", None, True


@dataclass
class Snippet:

    lines: list[str]

    def __getitem__(self, lineno: int) -> str:
        """Gets the line at the given 1-based line number."""
        if lineno > len(self.lines):
            return ""
        return self.lines[lineno - 1]

    def lines_between(self, start: int, last: int) -> Iterator[tuple[int, str]]:
        """Yields lines between the given 1-based line numbers (inclusive)."""
        start = max(start, 1)
        yield from enumerate(self.lines[start - 1 : last], start)


def calc_highlight_range(line: str, offset: int, substr: Any) -> tuple[int, int] | None:
    if substr is None:
        return  # early escape when substr value not given

    value = str(substr)
    if not value.strip():
        return  # prevent highlighting empty line

    try:
        col_start = line.index(value, offset)
    except ValueError:
        return

    col_end = col_start + len(value)
    return col_start, col_end



from __future__ import annotations

from typing import Literal

from pydantic import Field

from tugboat.schemas.basic import Empty
from tugboat.schemas.manifest import Manifest


class Spec(Empty):
    """This class is preserved for debugging purposes."""


class DebugManifest(Manifest):
    """
    Internal class for debugging purposes.

    This class could be constructed with the following manifest:

    .. code-block:: yaml

        apiVersion: tugboat.example.com/v1
        kind: Debug
        metadata:
          generateName: test-
        spec:
          # add any fields here

    :meta private:
    """

    apiVersion: Literal["tugboat.example.com/v1"]  # type: ignore[reportIncompatibleVariableOverride]
    kind: Literal["Debug"]  # type: ignore[reportIncompatibleVariableOverride]
    spec: Spec | None = Field(default_factory=Spec)



from __future__ import annotations

import contextlib
import typing
from typing import cast

from ruamel.yaml.comments import CommentedBase, CommentedMap, CommentedSeq
from ruamel.yaml.scalarstring import (
    DoubleQuotedScalarString,
    FoldedScalarString,
    LiteralScalarString,
    PlainScalarString,
    SingleQuotedScalarString,
)

from tugboat.types import Field

if typing.TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any

    from ruamel.yaml.mergevalue import MergeValue

    type IntTuple = tuple[int, ...]


def get_line_column(
    doc: CommentedMap, loc: Sequence[int | str], value: Any | Field | None
) -> tuple[int, int]:
    """
    Get the line and column number for a given location in the YAML document.

    Parameters
    ----------
    doc : CommentedMap
        The parsed YAML document.
    loc : Sequence[int | str]
        Path to the location in the document.
    value : Any | Field | None
        Value to locate a more specific position, if applicable. If a Field is
        provided, this function will attempt to find the position of the key in
        the map. Otherwise, it will search for the value in the field.

    Returns
    -------
    tuple[int, int]
        Line and column numbers (0-based).
    """
    # navigate through the path
    parent_node = None
    current_node = doc
    key = None
    fallback_position = (0, 0)
    assume_indent_size = 2

    for key in loc:
        parent_node = current_node

        try:
            current_node = current_node[key]  # type: ignore[reportIndexIssue]
        except (KeyError, IndexError, TypeError):
            # navigation failed, use fallback position from parent
            if isinstance(parent_node, CommentedBase):
                fallback_position = (parent_node.lc.line, parent_node.lc.col)
            break

        if isinstance(current_node, CommentedBase):
            fallback_position = (current_node.lc.line, current_node.lc.col)
            assume_indent_size = current_node.lc.col - parent_node.lc.col

        if not isinstance(current_node, CommentedSeq | CommentedMap):
            break  # prevent navigation further

    if key is None:
        return fallback_position

    # if the value is 'Field' type, return the position of the key
    if isinstance(value, Field):
        parent_node = cast("CommentedMap", parent_node)
        try:
            return parent_node.lc.key(value)
        except (KeyError, AttributeError):
            return fallback_position

    # for the rest of the cases, default to the start of the field value
    with contextlib.suppress(KeyError):
        fallback_position = get_value_linecol(parent_node, key)

    # when the value is a string, try to find its exact position
    if isinstance(current_node, str) and isinstance(value, str):
        if linecol := calculate_substring_linecol(
            parent_node=parent_node,
            current_node=current_node,
            key=key,
            substring=value,
            indent_size=assume_indent_size,
        ):
            return linecol

    return fallback_position


def is_anchor_node(parent_node: CommentedBase, key: int | str) -> bool:
    """
    Check if a child node is an anchor (&anchor).

    Returns
    -------
    bool
        True if this is an anchor node, False otherwise.
    """
    try:
        target_node = parent_node[key]  # type: ignore[reportIndexIssue]
    except (KeyError, IndexError, TypeError):
        return False

    anchor: Callable | None = getattr(target_node, "yaml_anchor", None)
    if anchor and anchor() and not is_alias_node(parent_node, key):
        return True

    return False


def is_alias_node(parent_node: CommentedBase, key: int | str) -> bool:
    """
    Check if a child node is an alias (*anchor).

    Returns
    -------
    bool
        True if this is an alias node, False otherwise.
    """
    if key is None:
        return False
    if not isinstance(parent_node, CommentedBase):
        return False

    if merge_values := getattr(parent_node, "_yaml_merge", None):
        merge_values = cast("MergeValue", merge_values)
        for value_dict in merge_values.value:
            if key in value_dict:
                return True

    try:
        child_lc = parent_node.lc.data[key]
    except (AttributeError, KeyError, IndexError):
        return False

    if isinstance(parent_node, dict):
        key_line, _, value_line, _ = child_lc
        return value_line < key_line

    elif isinstance(parent_node, list):
        value_line, _ = child_lc
        if value_line < parent_node.lc.line:
            return True

        if (prev_idx := cast("int", key) - 1) >= 0:
            prev_value_line, _ = parent_node.lc.data[prev_idx]
            if value_line < prev_value_line:
                return True

    return False


def get_value_linecol(parent_node: CommentedBase, key: int | str) -> tuple[int, int]:
    """Find the start of the field value. Returns 0-based line and column numbers."""
    if isinstance(parent_node, CommentedMap):
        key = cast("str", key)
        if is_alias_node(parent_node, key):
            line, col = parent_node.lc.key(key)
            # adjust to point after the key name and colon
            col += len(str(key)) + 2
            return (line, col)
        else:
            return parent_node.lc.value(key)

    elif isinstance(parent_node, CommentedSeq):
        key = cast("int", key)
        # NOTE when the value is an alias, ruamel.yaml returns the line/col of the anchor
        if not is_alias_node(parent_node, key):
            return parent_node.lc.key(key)

    raise KeyError


def calculate_substring_linecol(
    *,
    parent_node: CommentedBase,
    current_node: str,
    key: int | str,
    substring: str,
    indent_size: int,
) -> tuple[int, int] | None:
    """
    Calculate the line and column position of a substring within a scalar value.

    Parameters
    ----------
    parent_node : CommentedBase
        The parent node for enhanced positioning.
    current_node : str
        The current node containing the text.
    key : int | str
        The key in the parent node.
    substring: str
        The substring to find within the text.
    indent_size: int
        The assumed indentation size for the current node.

    Returns
    -------
    tuple[int, int] | None
        Line and column numbers (0-based), or None if substring not found or is alias.
    """
    # early exits
    if not substring:
        return None
    if is_alias_node(parent_node, key):
        return None
    if substring not in current_node:
        return None

    # dispatch based on scalar type
    if isinstance(current_node, LiteralScalarString):
        # literal block scalar (|)
        if isinstance(parent_node, CommentedSeq):
            return calculate_literal_substring_linecol_in_array(
                parent_node=cast("CommentedSeq", parent_node),
                current_node=current_node,
                key=cast("int", key),
                substring=substring,
            )
        if isinstance(parent_node, CommentedMap):
            return calculate_literal_substring_linecol_in_map(
                parent_node=cast("CommentedMap", parent_node),
                current_node=current_node,
                key=cast("str", key),
                substring=substring,
                indent_size=indent_size,
            )

    if isinstance(current_node, FoldedScalarString):
        # folded scalar (>) merges adjacent lines
        # so we can't tell the position of the substring reliably
        return None

    if isinstance(current_node, PlainScalarString):
        # plain scalar string type is found when the field contains anchor or alias
        # this breaks the logic of line/column calculation
        return None

    # default
    if linecol := calculate_simple_substring_linecol(
        parent_node=parent_node,
        current_node=current_node,
        key=key,
        substring=substring,
    ):
        return linecol

    return None


def calculate_literal_substring_linecol_in_map(
    *,
    parent_node: CommentedMap,
    current_node: LiteralScalarString,
    key: str,
    substring: str,
    indent_size: int,
) -> tuple[int, int]:
    """
    ```yaml
    foo: |-
      Lorem ipsum dolor sit amet,
      consectetur adipiscing elit.
    ```

    * key line/col points to the position of `foo:`
    * value line/col points to the position of indicator `|-`
    """
    _, key_col, value_line, _ = cast("IntTuple", parent_node.lc.data[key])

    idx_substring = current_node.find(substring)
    cnt_lines_before = current_node.count("\n", 0, idx_substring)
    idx_last_newline = current_node.rfind("\n", 0, idx_substring)
    offset_col = idx_substring - idx_last_newline - 1

    return (
        value_line + cnt_lines_before + 1,
        key_col + indent_size + offset_col,
    )


def calculate_literal_substring_linecol_in_array(
    *,
    parent_node: CommentedSeq,
    current_node: LiteralScalarString,
    key: int,
    substring: str,
) -> tuple[int, int]:
    """
    ```yaml
    - |-
      Lorem ipsum dolor sit amet,
      consectetur adipiscing elit.
    ```

    * key line/col points to the position of indicator `|-`
    """
    value_line, value_col = cast("IntTuple", parent_node.lc.key(key))

    idx_substring = current_node.find(substring)
    cnt_lines_before = current_node.count("\n", 0, idx_substring)
    idx_last_newline = current_node.rfind("\n", 0, idx_substring)
    offset_col = idx_substring - idx_last_newline - 1

    return (
        value_line + cnt_lines_before + 1,
        value_col + offset_col,
    )


def calculate_simple_substring_linecol(
    *,
    parent_node: CommentedBase,
    current_node: str,
    key: int | str,
    substring: str,
) -> tuple[int, int] | None:
    """
    Calculate substring line/col if the current node *looks like* a simple string.

    ```yaml
    foo: Lorem ipsum dolor sit amet
    ```
    """
    # find the start of the field value
    if isinstance(parent_node, CommentedMap):
        value_line, value_col = cast("IntTuple", parent_node.lc.value(key))
    elif isinstance(parent_node, CommentedSeq):
        value_line, value_col = cast("IntTuple", parent_node.lc.key(key))
    else:
        return None  # unknown parent node type

    # adjust for quoted strings
    additional_offset = 0
    if isinstance(current_node, DoubleQuotedScalarString | SingleQuotedScalarString):
        additional_offset = 1

    # calculate substring position
    idx_substring = current_node.find(substring)
    if current_node.find("\n", 0, idx_substring) != -1:
        # substring spans multiple lines
        return None

    return (
        value_line,
        value_col + additional_offset + idx_substring,
    )



from __future__ import annotations

import typing
from abc import ABC, abstractmethod

from pydantic import BaseModel, ConfigDict, Field

from tugboat._vendor.pygments.token import Whitespace

if typing.TYPE_CHECKING:
    from typing import Self

    from tugboat._vendor.pygments.token import _TokenType

    type Lexeme = tuple[int, _TokenType, str]


class Node(ABC, BaseModel):
    """Base class for AST nodes."""

    model_config = ConfigDict(frozen=True)

    children: list[Node] = Field(default_factory=list)

    @classmethod
    @abstractmethod
    def parse(cls, lexemes: list[Lexeme]) -> Node | None:
        """
        Parse a list of tokens into AST. This method may pop items from the list
        when it consumes them.

        Parameters
        ----------
        lexemes : list[Lexeme]
            The list of tokens to parse.

        Returns
        -------
        The AST node or None if parsing failed.
        """


class Unexpected(Node):
    """Node to represent unparsable tokens or syntax errors."""

    pos: int
    text: str
    msg: str | None = None

    @classmethod
    def parse(cls, lexemes: list[Lexeme]) -> Self | None:
        if not lexemes:
            return
        pos, _, text = lexemes.pop(0)
        return cls(pos=pos, text=text)

    @classmethod
    def create(cls, lexemes: list[Lexeme], msg: str | None = None) -> Self:
        start_pos, _, _ = lexemes[0]
        return cls(
            pos=start_pos,
            text="".join(text for _, _, text in lexemes),
            msg=msg,
        )


def next_lexeme_is(
    lexemes: list[Lexeme],
    token: _TokenType | None = None,
    text: str | None = None,
    shift: int = 0,
) -> bool:
    """Check if the next token in the list matches the given token or text."""
    if len(lexemes) <= 0:
        return False

    _, next_token, next_text = lexemes[shift]
    if token is not None and next_token not in token:
        return False
    if text is not None and next_text != text:
        return False
    return True


def consume_whitespaces(lexemes: list[Lexeme]) -> list[Lexeme]:
    """Pop all leading whitespace tokens from the lexemes stream."""
    buffer = []
    while next_lexeme_is(lexemes, Whitespace):
        buffer.append(lexemes.pop(0))
    return buffer



from __future__ import annotations

import os
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict

from tugboat.schemas.basic import Array
from tugboat.schemas.manifest import Manifest
from tugboat.schemas.workflow import WorkflowSpec

if os.getenv("DOCUTILSCONFIG"):
    __all__ = ["CronWorkflowSpec"]


class CronWorkflowSpec(BaseModel):
    """
    `CronWorkflowSpec`_ is the specification of a CronWorkflow.

    .. _CronWorkflowSpec: https://argo-workflows.readthedocs.io/en/latest/fields/#cronworkflowspec
    """

    model_config = ConfigDict(extra="forbid")

    concurrencyPolicy: str | None = None
    failedJobsHistoryLimit: int | None = None
    schedule: str | None = None
    schedules: Array[str] | None = None
    startingDeadlineSeconds: int | None = None
    successfulJobsHistoryLimit: int | None = None
    suspend: bool | None = None
    timezone: str | None = None
    when: str | None = None
    workflowSpec: WorkflowSpec

    stopStrategy: Any | None = None
    workflowMetadata: Any | None = None

    def __hash__(self):
        return hash((self.schedule, self.schedules, self.workflowSpec))


class CronWorkflow(Manifest[CronWorkflowSpec]):
    """
    `CronWorkflows`_ are workflows that run on a schedule.

    .. _CronWorkflows: https://argo-workflows.readthedocs.io/en/latest/cron-workflows/
    """

    apiVersion: Literal["argoproj.io/v1alpha1"]  # type: ignore[reportIncompatibleVariableOverride]
    kind: Literal["CronWorkflow"]  # type: ignore[reportIncompatibleVariableOverride]
    spec: CronWorkflowSpec



from __future__ import annotations

import collections
import sys

from tugboat.types import PathLike
from tugboat.utils import join_with_and


class Stdin(PathLike):
    """A path-like object that reads from stdin."""

    def __init__(self):
        super().__init__("<stdin>")

    def is_file(self) -> bool:
        return True

    def read_text(self) -> str:
        return sys.stdin.read()


class DiagnosesCounter(collections.Counter):

    def summary(self) -> str:
        parts = []
        if count := self["error"]:
            parts.append(f"{count} errors")
        if count := self["failure"]:
            parts.append(f"{count} failures")
        if count := self["warning"]:
            parts.append(f"{count} warnings")

        if parts:
            summary = join_with_and(parts, quote=False, sort=False)
            return f"Found {summary}"

        return "All passed!"

    def has_any_error(self) -> bool:
        return any((self["error"], self["failure"]))



from __future__ import annotations

from pydantic import BaseModel, ConfigDict

from tugboat.schemas.basic import ConfigKeySelector


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)


class EnvVar(_BaseModel):
    """
    `EnvVar`_ represents an environment variable present in a Container.

    .. _EnvVar: https://argo-workflows.readthedocs.io/en/latest/fields/#envvar
    """

    name: str
    value: str | None = None
    valueFrom: EnvVarSource | None = None


class EnvVarSource(_BaseModel):
    configMapKeyRef: ConfigKeySelector | None = None
    fieldRef: ObjectFieldSelector | None = None
    resourceFieldRef: ResourceFieldSelector | None = None
    secretKeyRef: ConfigKeySelector | None = None


class ObjectFieldSelector(_BaseModel):
    apiVersion: str | None = None
    fieldPath: str


class ResourceFieldSelector(_BaseModel):
    containerName: str | None = None
    divisor: str | None = None
    resource: str


class EnvFromSource(_BaseModel):
    """
    `EnvFromSource`_ represents the source of a set of ConfigMaps.

    .. _EnvFromSource: https://argo-workflows.readthedocs.io/en/latest/fields/#envfromsource
    """

    configMapRef: OptionalName | None = None
    prefix: str | None = None
    secretRef: OptionalName | None = None


class OptionalName(_BaseModel):
    """
    Represents a reference to a ConfigMap or Secret.
    This class is utilized by both the `ConfigMapEnvSource`_ and `SecretEnvSource`_ classes.

    .. _ConfigMapEnvSource:
       https://argo-workflows.readthedocs.io/en/latest/fields/#configmapenvsource
    .. _SecretEnvSource:
       https://argo-workflows.readthedocs.io/en/latest/fields/#secretenvsource
    """

    name: str
    optional: bool | None = None
